/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: children.c,v 1.4 1994/03/11 14:34:52 carlsson Exp $
 * $Log: children.c,v $
 * Revision 1.4  1994/03/11  14:34:52  carlsson
 * Changed the signal that is sent to the children from `stop_children()'
 * from SIGQUIT to SIGKILL.
 *
 * Revision 1.3  1994/02/28  18:54:51  carlsson
 * #ifdef VERBOSE introduced.
 *
 * Revision 1.2  1993/12/21  10:41:42  carlsson
 * Initial revision
 */

#include <stdio.h>
#include <sys/wait.h>
#include <sys/time.h>

#include "children.h"
#include "system.h"

#define ERR_CHILD_MAX -2
#define ERR_CHILD_NO_INIT -1
#define CHILD_LOAD_MANAGER 0
#define CHILD_REXEC_MANAGER 1
#define CHILD_MAX 2

#define TIMELIMIT 5

typedef struct {
	int pid;
	char name[MAX_PATH_LEN];
	struct timeval timestamp;
	int startable;
} child;

static int global_child_count = 0;
static int global_is_init = FALSE;
static child global_child_tab[CHILD_MAX];

static void (*global_callback_fn)(void) = NULL;

extern int wait3(int *, int, struct rusage *);
extern int fork();
extern int execlp();
extern int setpgrp(int, int);
extern pid_t setsid();
extern int sigblock(int);
extern int sigsetmask(int);
extern int gettimeofday(struct timeval *, struct timezone *);
extern int fprintf();
extern int strncpy(char *, char *, int);


static void restart_child(int pid)
{
	int i;
	struct timeval timestamp;
	
	for( i = 0; i < global_child_count; i++ ) {
		if( global_child_tab[i].pid == pid ) {
			gettimeofday(&timestamp, (struct timezone *)NULL);

			if( timestamp.tv_sec - global_child_tab[i].timestamp.tv_sec < TIMELIMIT ) {
#ifdef VERBOSE
				fprintf(stderr, "Can't start %s\n", global_child_tab[i].name);
#endif
				syslog(LOG_WARNING, 
                            "Load-Daemon: cannot start `%s'.",
				       global_child_tab[i].name);

				global_child_tab[i].startable = FALSE;
			}
			else {
				global_child_tab[i].timestamp = timestamp;

				if( (global_child_tab[i].pid = fork()) == 0 ) {
					if( global_callback_fn ) {
						(*global_callback_fn)();
					}
					sigsetmask(0);
					signal(SIGCHLD, SIG_DFL);

#ifdef VERBOSE
					fprintf(stderr, "Restarting %s\n", global_child_tab[i].name);
#endif
					syslog(LOG_INFO, 
                            "Load-Daemon: restarting `%s'...",
					       global_child_tab[i].name);
				
					execlp(global_child_tab[i].name, global_child_tab[i].name, (char *)NULL);
					syslog(LOG_ERR, 
		       "Load-Daemon: cannot execute `%s'; reason: %m",
					       global_child_tab[i].name);
					exit(-1);
				}
			}
		}
	}
}


static void handle_SIGCHLD (void)
{
	union wait status;               /* Status of the child process */
	int pid;                         /* Process-id of child process */
	struct rusage resources;
#if 1
	/* XXX (schoenfr): don't block, boy: */
	while((pid = wait3((int *) &status, WNOHANG | WUNTRACED, 
			   (struct rusage *) &resources)) < 0
	      && errno == EINTR)
	    continue;
#else
	while( ((pid = wait3((int *)&status, 0, (struct rusage *) &resources)) < 0) && (errno == EINTR) ) {
		;
	}
#endif
  
	if( (!WIFEXITED(status) && !WIFSIGNALED(status)) || pid == 0 ) {
		return;
	}

	restart_child(pid);
}
			


void init_children(void (*in_callback_fn)(void))
{
	global_callback_fn = in_callback_fn;
	signal(SIGCHLD, handle_SIGCHLD);
	global_is_init = TRUE;
}



void kill_child(int in_child, int in_sig)
{
	if( in_child < global_child_count && in_child >= 0 ) {
		kill(global_child_tab[in_child].pid, in_sig);
	}
}

void stop_children(void)
{
	int i;

	signal(SIGCHLD, SIG_DFL);
	for( i = 0; i < global_child_count; i++ ) {
		if( global_child_tab[i].startable ) {
			kill(global_child_tab[i].pid, SIGKILL);
		}
	}

	global_is_init = FALSE;
	global_child_count = 0;
	global_callback_fn = NULL;
}

void restart_children(void)
{
	int i;
	
	for( i = 0; i < global_child_count; i++ ) {
		if( !global_child_tab[i].startable ) {
			global_child_tab[i].startable = TRUE;
			restart_child(global_child_tab[i].pid);
		}
	}
}


int start_child(char *in_name)
{
	int mask;

	if( !global_is_init ) {
		return( ERR_CHILD_NO_INIT );
	}
		
	if( global_child_count >= CHILD_MAX ) {
		return( ERR_CHILD_MAX );
	}


	strncpy(global_child_tab[global_child_count].name, in_name, MAX_PATH_LEN);
	gettimeofday(&global_child_tab[global_child_count].timestamp, (struct timezone *)NULL);
	global_child_tab[global_child_count].startable = TRUE;
	
	mask = sigblock(sigmask(SIGCHLD));
	global_child_count++;
	
	if( (global_child_tab[global_child_count - 1].pid = fork()) == 0 ) {
		if( global_callback_fn ) {
			(*global_callback_fn)();
		}
		sigsetmask(mask);
		signal(SIGCHLD, SIG_DFL);

#ifdef VERBOSE
		fprintf(stderr, "Starting %s\n", in_name);
#endif
		
		execlp(in_name, in_name, (char *)NULL);
		fprintf(stderr, "Execlp error\n");
		exit(-1);
	}

	sigsetmask(mask);

	return( global_child_count - 1 );
}

